
#include <windows.h>
#include <winsock.h>

#include "amxxmodule.h"

#pragma comment( lib , "user32.lib" ) 
#pragma comment( lib , "ws2_32.lib" ) 

#define WM_SOCKET ( WM_USER + 108 )
#define SOCKET_TCP 1
#define SOCKET_UDP 2

HINSTANCE hInstance = 0;
HWND hWinsockWindow = 0;
int iFwdConnected = -1;
int iFwdDisconnected = -1;
int iFwdConnectionRequest = -1;
int iFwdDataArrival = -1;

LRESULT CALLBACK WinProc( HWND hWnd , UINT message , WPARAM wParam , LPARAM lParam );

// native winsock_connect(_hostname[],_port,_protocol=SOCKET_TCP,&_error)
static cell AMX_NATIVE_CALL winsock_connect( AMX *amx , cell *params )
{
	cell *err = MF_GetAmxAddr( amx , params[ 4 ] );
	*err = 0;

	if ( !hWinsockWindow )
	{
		*err = 1;
		return -1;
	}

	int len;
	char *hostname = MF_GetAmxString( amx , params[ 1 ] , 0 , &len );

	if( len == 0 ) 
	{
		*err = 2;
		return -1;
	}

	SOCKET sSocket = socket( AF_INET , ( params[ 3 ] == SOCKET_TCP ) ? SOCK_STREAM : SOCK_DGRAM , 0 );

	if ( sSocket == INVALID_SOCKET ) 
	{
		*err = WSAGetLastError();
		return -1;
	}
	
	if ( WSAAsyncSelect( sSocket , hWinsockWindow , WM_SOCKET , ( FD_CONNECT | FD_CLOSE | FD_READ ) ) == SOCKET_ERROR )
	{
		*err = WSAGetLastError();
		return -1;
	}

	struct sockaddr_in server;
	struct hostent *host_info;
	unsigned long addr;
	memset( &server , 0 , sizeof( server ) );

	if ( ( addr = inet_addr( hostname ) ) != INADDR_NONE ) 
	{
		memcpy( (char *)&server.sin_addr , &addr , sizeof( addr ) );
	}
	else 
	{
		host_info = gethostbyname( hostname );
		if ( host_info == NULL ) 
		{
			*err = WSAGetLastError();
			return -1;
		}
		memcpy( (char *)&server.sin_addr , host_info->h_addr , host_info->h_length );
	}

	server.sin_family = AF_INET;
	server.sin_port = htons( (unsigned int)params[ 2 ] );
	
	if ( ( connect( sSocket , (struct sockaddr*)&server , sizeof( server ) ) == SOCKET_ERROR ) && ( WSAGetLastError() != WSAEWOULDBLOCK ) ) 
	{
		*err = WSAGetLastError();
		return -1;
	}

	return sSocket;
}

// native winsock_listen(_port,_protocol = SOCKET_TCP,&_error)
static cell AMX_NATIVE_CALL winsock_listen( AMX *amx , cell *params )
{
	cell *err = MF_GetAmxAddr( amx , params[ 3 ] );
	*err = 0;

	if ( !hWinsockWindow )
	{
		*err = 1;
		return -1;
	}

	SOCKET sSocket = socket( AF_INET , ( params[ 2 ] == SOCKET_TCP ) ? SOCK_STREAM : SOCK_DGRAM , 0 );

	if ( sSocket == INVALID_SOCKET ) 
	{
		*err = WSAGetLastError();
		return -1;
	}

	int bOptVal = TRUE;
	if ( setsockopt( sSocket , SOL_SOCKET , SO_REUSEADDR , (char*)&bOptVal , sizeof( bOptVal ) ) == SOCKET_ERROR )
	{
		*err = WSAGetLastError();
		return -1;
	}

	if ( WSAAsyncSelect( sSocket , hWinsockWindow , WM_SOCKET , ( FD_ACCEPT | FD_CLOSE | FD_READ ) ) == SOCKET_ERROR )
	{
		*err = WSAGetLastError();
		return -1;
	}

	struct sockaddr_in server;
	memset( &server , 0 , sizeof( server ) );

	server.sin_family = AF_INET;
	server.sin_addr.s_addr = INADDR_ANY;
	server.sin_port = htons( (unsigned int)params[ 1 ] );

	if ( ( bind( sSocket , (SOCKADDR *)&server , sizeof( server ) ) == SOCKET_ERROR ) && ( WSAGetLastError() != WSAEWOULDBLOCK ) ) 
	{
		*err = WSAGetLastError();
		return -1;
	}

	if ( ( listen( sSocket , 1 ) == SOCKET_ERROR ) && ( WSAGetLastError() != WSAEWOULDBLOCK ) ) 
	{
		*err = WSAGetLastError();
		return -1;
	}

	return sSocket;
}

// native winsock_accept(_socket,&_error)
static cell AMX_NATIVE_CALL winsock_accept( AMX *amx , cell *params )
{
	cell *err = MF_GetAmxAddr( amx , params[ 2 ] );
	*err = 0;

	SOCKET sSocket = accept( params[ 1 ] , NULL , NULL );

	if ( ( sSocket == INVALID_SOCKET ) && ( WSAGetLastError() != WSAEWOULDBLOCK ) )
	{
		*err = WSAGetLastError();
		return -1;
	}

	return sSocket;
}

// native winsock_close(_socket)
static cell AMX_NATIVE_CALL winsock_close(AMX *amx, cell *params)
{
	return ( closesocket( params[ 1 ] ) == SOCKET_ERROR ) ? WSAGetLastError() : 0;
}

// native winsock_recv(_socket,_data[],_length,&_error)
static cell AMX_NATIVE_CALL winsock_recv( AMX *amx , cell *params )
{
	cell *err = MF_GetAmxAddr( amx , params[ 4 ] );
	int tmp = -1;
	*err = 0;

	int length = params[ 3 ];

	char *tmpchar = new char[ length ];
	if( tmpchar == NULL )
	{
		*err = 1;
		return -1;
	}

	memset( tmpchar , 0 , length );
	if ( ( ( tmp = recv( params[ 1 ] , tmpchar , length - 1 , 0 ) ) == SOCKET_ERROR ) && ( WSAGetLastError() != WSAEWOULDBLOCK ) )
	{
		delete [] tmpchar;
		*err = WSAGetLastError();
		return -1;
	}

	tmpchar[ tmp ] = '\0';

	int nlen = 0;
	int max = length - 1;
	const char* src = tmpchar;
	cell* dest = MF_GetAmxAddr( amx , params[ 2 ] );

	while ( max-- && ( nlen < tmp ) )
	{
		*dest++ = (cell)*src++;
		nlen++;
	}

	*dest = 0;
	delete [] tmpchar;

	return tmp;
}

// native winsock_bytesavailable(_socket,&_error)
static cell AMX_NATIVE_CALL winsock_bytesavailable( AMX *amx , cell *params )
{
	cell *err = MF_GetAmxAddr( amx , params[ 2 ] );
	u_long iNumBytes = 0;

	if ( ioctlsocket( params[ 1 ] , FIONREAD , &iNumBytes ) == SOCKET_ERROR )
	{
		*err = WSAGetLastError();
		 return -1;
	}

	return iNumBytes;
}

// native winsock_send(_socket,_data[],_length,&_error)
static cell AMX_NATIVE_CALL winsock_send( AMX *amx , cell *params )
{
	int len = params[ 3 ];
	char *data = MF_GetAmxString( amx , params[ 2 ] , 0 , &len );
	cell *err = MF_GetAmxAddr( amx , params[ 4 ] );

	int iRetVal = send( params[ 1 ] , data , len , 0 );
	*err = 0;

	if ( ( iRetVal == SOCKET_ERROR ) && ( WSAGetLastError() != WSAEWOULDBLOCK ) ) 
	{
		*err = WSAGetLastError();
		 return -1;
	}
	
	return iRetVal;
}

static char *g_buffer = NULL;
static size_t g_buflen = 0;
// native winsock_send2(_socket,_data[],_length,&_error)
static cell AMX_NATIVE_CALL winsock_send2( AMX *amx , cell *params )
{
	int len = params[ 3 ];
	cell *err = MF_GetAmxAddr( amx , params[ 4 ] );
	*err = 0;

	if ( (size_t)len > g_buflen )
	{
		delete [] g_buffer;
		g_buffer = new char[ len + 1 ];
		g_buflen = len;
	}

	cell *pData = MF_GetAmxAddr( amx , params[ 2 ] );
	char *pBuffer = g_buffer;

	while ( len-- )
		*pBuffer++ = (char)*pData++;

	int iRetVal = send( params[ 1 ] , g_buffer , len , 0 );

	if ( ( iRetVal == SOCKET_ERROR ) && ( WSAGetLastError() != WSAEWOULDBLOCK ) ) 
	{
		*err = WSAGetLastError();
		 return -1;
	}

	return iRetVal;
}

// native winsock_getpeeraddr(_socket,_dest[],_max,&_port,&_error)
static cell AMX_NATIVE_CALL winsock_getpeeraddr( AMX *amx , cell *params )
{
	cell *err = MF_GetAmxAddr( amx , params[ 5 ] );
	*err = 0;

	struct sockaddr_in saPeerAddr;
	int iPeerLen = sizeof( saPeerAddr );

	if ( getpeername( params[ 1 ] , (struct sockaddr*)&saPeerAddr , &iPeerLen ) == SOCKET_ERROR )
	{
		*err = WSAGetLastError();
		return -1;
	}

	cell *dest = MF_GetAmxAddr( amx , params[ 2 ] );
	cell *max = MF_GetAmxAddr( amx , params[ 3 ] );
	cell *port = MF_GetAmxAddr( amx , params[ 4 ] );

	*port = ntohs( saPeerAddr.sin_port );
	const char *tmp = (const char*)inet_ntoa( saPeerAddr.sin_addr );
	int addrlen = strlen( tmp );
	int nlen = 0;

	while ( ( nlen < addrlen ) && ( nlen < (int)max ) )
	{
		*dest++ = (cell)*tmp++;
		nlen++;
	}

	return nlen;
}

int Create_Window()
{
	WNDCLASSEX wClass;
	ZeroMemory( &wClass , sizeof( WNDCLASSEX ) );
	wClass.cbClsExtra=NULL;
	wClass.cbSize=sizeof( WNDCLASSEX );
	wClass.cbWndExtra=NULL;
	wClass.hbrBackground=(HBRUSH)COLOR_WINDOW;
	wClass.hCursor=LoadCursor( NULL , IDC_ARROW );
	wClass.hIcon=NULL;
	wClass.hIconSm=NULL;
	wClass.hInstance=hInstance;
	wClass.lpfnWndProc=(WNDPROC)WinProc;
	wClass.lpszClassName="WinsockClass";
	wClass.lpszMenuName=NULL;
	wClass.style=CS_HREDRAW|CS_VREDRAW;

	if ( !RegisterClassEx( &wClass ) )
		return 0;

	hWinsockWindow = CreateWindowEx(NULL,
				"WinsockClass",
				"Winsock Module",
				WS_OVERLAPPEDWINDOW,
				50,
				50,
				50,
				50,
				NULL,
				NULL,
				hInstance,
				NULL);

	if ( !hWinsockWindow )
		return 0;

	//For testing
	//ShowWindow( hWinsockWindow , 1 );

	return 1;
}

LRESULT CALLBACK WinProc( HWND hWnd , UINT msg , WPARAM wParam , LPARAM lParam )
{
	if ( msg == WM_SOCKET )
	{
		int iErrorCode = WSAGETSELECTERROR( lParam );

		switch ( WSAGETSELECTEVENT( lParam ) )
		{
			case FD_CONNECT:
			{
				MF_ExecuteForward( iFwdConnected , wParam , iErrorCode );
			}
			break;
			
			case FD_ACCEPT:
			{	
				MF_ExecuteForward( iFwdConnectionRequest , wParam , iErrorCode );
			}
			break;

			case FD_READ:
			{
				MF_ExecuteForward( iFwdDataArrival , wParam , iErrorCode );
			}
			break;

			case FD_CLOSE:
			{
				MF_ExecuteForward( iFwdDisconnected , wParam , iErrorCode );
			}
		}

	}

	return DefWindowProc( hWnd , msg , wParam , lParam );
}

AMX_NATIVE_INFO winsock_natives[] = 
{
	{"Winsock_Connect",	   winsock_connect},
	{"Winsock_Close",	   winsock_close},
	{"Winsock_Listen",	   winsock_listen},
	{"Winsock_Accept",	   winsock_accept},
	{"Winsock_RecvData",	   winsock_recv},
	{"Winsock_SendText",	   winsock_send},
	{"Winsock_SendBinary",	   winsock_send2},
	{"Winsock_GetPeerAddr",	   winsock_getpeeraddr},
	{"Winsock_BytesAvailable", winsock_bytesavailable},
	{NULL,NULL}
};

void OnAmxxAttach()
{
	hInstance = (HINSTANCE)GetModuleHandle( NULL );
	Create_Window();
	WSADATA WsaDat;
	if ( !WSAStartup( MAKEWORD( 2 , 2 ) , &WsaDat ) )
	{
		MF_AddNatives( winsock_natives );
	}
}

void OnPluginsLoaded()
{
	iFwdConnected = MF_RegisterForward( "Winsock_Connected" , ET_IGNORE , FP_CELL , FP_CELL , FP_DONE );
	iFwdDisconnected = MF_RegisterForward( "Winsock_Disconnected" , ET_IGNORE , FP_CELL , FP_CELL , FP_DONE );
	iFwdConnectionRequest = MF_RegisterForward( "Winsock_ConnectionRequest" , ET_IGNORE , FP_CELL , FP_CELL , FP_DONE );
	iFwdDataArrival = MF_RegisterForward( "Winsock_DataArrival" , ET_IGNORE , FP_CELL , FP_CELL , FP_DONE );
}

void OnAmxxDetach()
{
	WSACleanup();
	DestroyWindow( hWinsockWindow );
	hWinsockWindow = 0;
	UnregisterClass( "WinsockClass" , hInstance );
	hInstance = 0;
	delete [] g_buffer;
}
